#include "qvminspector.h"

qvminspector_t::qvminspector_t( qwidget_t *parent ) : qmain_window_t( parent ), file_header( nullptr ), vmctx( nullptr )
{
    ui.setupUi( this );
    ui.virt_instrs->setColumnWidth( 0, 180 );
    ui.virt_instrs->setColumnWidth( 1, 150 );
    ui.virt_instrs->setColumnWidth( 2, 190 );
    ui.virt_instrs->setColumnWidth( 3, 200 );

    connect( ui.action_open, &QAction::triggered, this, &qvminspector_t::on_open );
    connect( ui.action_close, &QAction::triggered, this, &qvminspector_t::on_close );
}

void qvminspector_t::on_close()
{
    exit( 0 );
}

void qvminspector_t::on_open()
{
    if ( file_header )
        free( file_header );

    file_header = nullptr;
    image_base = 0u, vm_entry_rva = 0u, module_base = 0u;

    file_path = QFileDialog::getOpenFileName(
        this, tr( "open vmp2 file" ), std::filesystem::current_path().string().c_str(), tr( "vmp2 file (*.vmp2)" ) );

    const auto &_file_path = file_path.toStdString();

    if ( file_path.isEmpty() )
    {
        dbg_msg( "invalid vmp2 file... no file selected..." );
        return;
    }

    if ( !std::filesystem::exists( _file_path ) )
    {
        dbg_msg( "vmp2 file does not exist..." );
        return;
    }

    const auto file_size = std::filesystem::file_size( _file_path );

    if ( !file_size )
    {
        dbg_msg( "invalid vmp2 file size..." );
        return;
    }

    qfile_t open_file( file_path );
    file_header = reinterpret_cast< vmp2::v4::file_header * >( malloc( file_size ) );

    if ( !open_file.open( QIODevice::ReadOnly ) )
    {
        dbg_msg( "failed to open vmp2 file..." );
        return;
    }

    memcpy( file_header, open_file.readAll().data(), file_size );

    if ( !init_data() )
    {
        dbg_msg( "failed to init vmp2 file data..." );
        return;
    }
}

void qvminspector_t::dbg_print( qstring_t dbg_output )
{
    ui.dbg_output_window->appendPlainText( dbg_output );
}

void qvminspector_t::dbg_msg( qstring_t dbg_output )
{
    qmsg_box_t msg_box;
    msg_box.setText( dbg_output );
    msg_box.exec();
    dbg_print( dbg_output );
}

bool qvminspector_t::init_data()
{
    if ( file_header->magic != VMP_MAGIC )
    {
        dbg_msg( "invalid magic bytes for vmp2 file..." );

        return false;
    }

    dbg_print( "valid magic bytes for vmp2 file..." );

    if ( file_header->version != vmp2::version_t::v4 )
    {
        dbg_msg( "invalid vmp2 file version... "
                 "this vminspector is compiled for version 4...\n" );

        return false;
    }

    vm_entry_rva = file_header->vm_entry_rva;
    image_base = file_header->image_base;
    image_size = file_header->module_size;

    module_base = reinterpret_cast< std::uintptr_t >( file_header ) + file_header->module_offset;
    return true;
}

void qvminspector_t::add_branch_children( qtree_widget_item_t *item, std::uintptr_t branch_addr )
{

}

void qvminspector_t::update_ui()
{

}

bool qvminspector_t::serialize_vmp2( std::vector< rtn_data_t > &virt_rtns, std::vector< std::uint8_t > &vmp2file )
{
    const auto file_header = reinterpret_cast< vmp2::v4::file_header * >( vmp2file.data() );

    if ( file_header->version != vmp2::version_t::v4 )
    {
        std::printf( "[!] invalid vmp2 file version... this build uses v3...\n" );
        return false;
    }

    auto first_rtn = reinterpret_cast< vmp2::v4::rtn_t * >( reinterpret_cast< std::uintptr_t >( file_header ) +
                                                            file_header->rtn_offset );

    for ( auto [ rtn_block, rtn_idx ] = std::pair{ first_rtn, 0ull }; rtn_idx < file_header->rtn_count;
          ++rtn_idx, rtn_block = reinterpret_cast< vmp2::v4::rtn_t * >(
                         reinterpret_cast< std::uintptr_t >( rtn_block ) + rtn_block->size ) )
    {
        virt_rtns.push_back( { rtn_block->vm_enter_offset, {} } );
        for ( auto [ code_block, block_idx ] = std::pair{ &rtn_block->code_blocks[ 0 ], 0ull };
              block_idx < rtn_block->code_block_count;
              ++block_idx, code_block = reinterpret_cast< vmp2::v4::code_block_t * >(
                               reinterpret_cast< std::uintptr_t >( code_block ) + code_block->next_block_offset ) )
        {
            auto block_vinstrs = reinterpret_cast< vm::instrs::virt_instr_t * >(
                reinterpret_cast< std::uintptr_t >( code_block ) + sizeof vmp2::v4::code_block_t +
                ( code_block->num_block_addrs * 8 ) );

            vm::instrs::code_block_t _code_block{ code_block->vip_begin };
            _code_block.jcc.has_jcc = code_block->has_jcc;
            _code_block.jcc.type = code_block->jcc_type;

            for ( auto idx = 0u; idx < code_block->num_block_addrs; ++idx )
                _code_block.jcc.block_addr.push_back( code_block->branch_addr[ idx ] );

            for ( auto idx = 0u; idx < code_block->vinstr_count; ++idx )
                _code_block.vinstrs.push_back( block_vinstrs[ idx ] );

            virt_rtns.back().rtn_blks.push_back( _code_block );
        }
    }

    return true;
}